//////////////////////////////////////////////////////////////////////////////////
//                                                                              //
//  This is a SocketCAN demo converted from C to Pascal by Zeljko Avramovic.    //
//  Original example has GNU General Public License v3.0 and can be found at    //
//  https://github.com/craigpeacock/CAN-Examples                                //
//                                                                              //
//  Virtual CAN adapter vcan0 is hard coded and you can bring it up like this:  //
//  sudo modprobe vcan                                                          //
//  sudo ip link add dev vcan0 type vcan                                        //
//  sudo ip link set vcan0 mtu 72            # needed for CAN FD                //
//  sudo ip link set vcan0 up                                                   //
//                                                                              //
//////////////////////////////////////////////////////////////////////////////////

program hlcanfdreceive; // higher level pascalized version of canfdreceive demo

{$mode delphi} {$H+}

uses
  sysutils, sockets, baseunix, can, can.hl, can.raw;

procedure perror(const S : string);
begin
  WriteLn(S, ', SocketError = ', SocketError);
end;

var
  i, BytesNo, Err: integer;
  CanHandle:       TCanHandle;
  CanAddress:      TCanSockAddress;
  CanInterface:    TCanIFreq;
  CanFDFrame:      TCanFDFrame;
begin
  try
    WriteLn('CAN FD Sockets Receive Demo');
    CanHandle := CanSocket(SOCK_RAW, CAN_RAW);
    if CanHandle < 0 then
    begin
      perror('Socket');
      Exit;
    end;

    CanInterface.Name := 'vcan0';
    if CanIOCtl(CanHandle, SIOCGIFINDEX, CanInterface) < 0 then
    begin
      perror('IOctl');
      Exit;
    end;

    CanAddress.Clear;
    CanAddress.CanFamily  := AF_CAN;
    CanAddress.CanIfIndex := CanInterface.IfIndex;

    if CanBind(CanHandle, CanAddress) < 0 then
    begin
      perror('Bind');
      Exit;
    end;

    Err := CanSetSockOpt(CanHandle, SOL_CAN_RAW, CAN_RAW_FD_FRAMES, ENABLE_CAN_OPTION, SizeOf(ENABLE_CAN_OPTION));
    if Err <> 0 then
    begin
      if Err = ESysENOPROTOOPT then
        WriteLn('CAN FD Protocol not available. Try newer kernel.');
      perror('SetSockopt');
      Exit;
    end;

    BytesNo := CanRead(CanHandle, CanFDFrame);
    if BytesNo < 0 then
    begin
      perror('Read: Error = ' + CanGetErrNo.ToString);
      Exit;
    end;

    if BytesNo = CANFD_MTU then    // test CAN FD with 'cansend vcan0 123##1223344556677889900'
      WriteLn('CAN FD frame received with length ', IntToStr(CanFDFrame.Length))
    else if BytesNo = CAN_MTU then // test legacy CAN with 'cansend vcan0 456#DEADBEEF'
      WriteLn('CAN legacy frame received with length ', IntToStr(CanFDFrame.Length))
    else
      WriteLn('Invalid CAN(FD) frame received');

    Write(format('0x%03X [%d] ', [CanFDFrame.ID.Value, CanFDFrame.Length]));
    for i := 0 to CanFDFrame.Length - 1 do
      Write(Format('%02.2X ', [CanFDFrame.Data[i]]));
    WriteLn('');

    if CanClose(CanHandle) < 0 then
    begin
      perror('Close');
      Exit;
    end;

  except
    on e: Exception do
      WriteLn(e.Message);
  end;
end.

